/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __houdini_SurfaceAccessorHoudini_h__
#define __houdini_SurfaceAccessorHoudini_h__

#define	FE_SAH_ASSERT_BOUNDS	TRUE

namespace fe
{

// TODO rework select() methodology to be reentrant

//* look at VM/VM_Math for simd

//* thread-safe only on isolated pages
//* different threads can't write to the same page

//* GA_AttributeRefMap could replace GEO_AttributeHandleList

/**************************************************************************//**
    @brief Houdini Surface Editing Implementation

	Note that attempts to access non-existant attributes isn't a failure.
	New attributes may be added when first written to by a set() method.

	@ingroup houdini
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessorHoudini:
	public SurfaceAccessorBase
{
	public:
						SurfaceAccessorHoudini(void):
							m_pGdp(NULL),
							m_pGdpChangeable(NULL),
							m_pPrimitive(NULL),
							m_pGroup(NULL)
						{
#ifdef FE_HOUDINI_USE_GA
							m_pIterator=NULL;
#else
							m_pPoint=NULL;
							m_pPointChangeable=NULL;
#endif
							setName("SurfaceAccessorHoudini");
						}

virtual					~SurfaceAccessorHoudini(void)
						{
#ifdef FE_HOUDINI_USE_GA
							delete m_pIterator;
#endif
#if FE_HOUDINI_HARDENING
							if(m_spHardening.isValid())
							{
								FEASSERT(m_spSurfaceAccessibleI.isValid());
								m_spSurfaceAccessibleI->lock();
								m_spHardening=NULL;
								m_spSurfaceAccessibleI->unlock();
							}
#endif
						}

		BWORD			bind(SurfaceAccessibleI::Element a_element,
								SurfaceAccessibleI::Attribute a_attribute)
						{
							m_attribute=a_attribute;

							String name;
							switch(a_attribute)
							{
								case SurfaceAccessibleI::e_generic:
								case SurfaceAccessibleI::e_position:
									name="P";
									break;
								case SurfaceAccessibleI::e_normal:
									name="N";
									break;
								case SurfaceAccessibleI::e_vertices:
									m_attrName="vertices";
									FEASSERT(a_element==
											SurfaceAccessibleI::e_primitive);
									m_element=a_element;
									return TRUE;
								case SurfaceAccessibleI::e_properties:
									m_attrName="properties";
									FEASSERT(a_element==
											SurfaceAccessibleI::e_primitive);
									m_element=a_element;
									return TRUE;
							}
							return bindInternal(a_element,name);
						}
		BWORD			bind(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							m_attribute=SurfaceAccessibleI::e_generic;
							return bindInternal(a_element,a_name);
						}

						//* as SurfaceAccessorI
virtual	String			type(void) const;

virtual	U32				count(void) const
						{
							FEASSERT(m_pGdp);

#if FALSE
							if(m_attribute!=SurfaceAccessibleI::e_generic &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									m_attribute!=
									SurfaceAccessibleI::e_properties &&
									!isBound())
							{
								return 0;
							}
#else
							if(m_attribute==SurfaceAccessibleI::e_position &&
									!isBound())
							{
								return 0;
							}
#endif

							switch(m_element)
							{
							case SurfaceAccessibleI::e_point:
#ifdef FE_HOUDINI_USE_GA
								return m_pGdp->getNumPoints();
#else
								return m_pGdp->points().entries();
#endif

//							case SurfaceAccessibleI::e_vertex:
//								return 0;

							case SurfaceAccessibleI::e_primitive:
#ifdef FE_HOUDINI_USE_GA
								return m_pGdp->getNumPrimitives();
#else
								return m_pGdp->primitives().entries();
#endif

							case SurfaceAccessibleI::e_detail:
								return 1;

							case SurfaceAccessibleI::e_pointGroup:
							case SurfaceAccessibleI::e_primitiveGroup:
								return m_pGroup? m_pGroup->entries(): 0;

							default:
								;
							}
							return 0;
						}
virtual	U32				subCount(U32 a_index) const
						{
							if(m_attribute!=SurfaceAccessibleI::e_generic &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									m_attribute!=
									SurfaceAccessibleI::e_properties &&
									!isBound())
							{
								return 0;
							}

							const_cast<SurfaceAccessorHoudini*>(this)->select(
									a_index);

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								FEASSERT(m_pPrimitive);
								return m_pPrimitive->getVertexCount();
							}

							return 1;

							// TODO check what really needs sub-indexes
//							return m_handle.entries();
						}


virtual	void			set(U32 a_index,U32 a_subIndex,String a_string)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(a_index<count());
							FEASSERT(a_subIndex<subCount(a_index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									m_attribute!=
									SurfaceAccessibleI::e_properties &&
									!isBound())
							{
								addString();
							}
							select(a_index);

							UT_String utString(a_string.c_str());
							m_handle.setString(utString,a_subIndex);
						}
virtual	String			string(U32 a_index,U32 a_subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(a_index<count());
							FEASSERT(a_subIndex<subCount(a_index));
#endif
							select(a_index);

							if(m_attribute==SurfaceAccessibleI::e_vertices ||
									m_attribute==
									SurfaceAccessibleI::e_properties)
							{
								return String();
							}

							UT_String utString;
							m_handle.getString(utString,a_subIndex);
							return String(utString.buffer());
						}

virtual	void			set(U32 a_index,U32 a_subIndex,I32 a_integer)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(a_index<count());
							FEASSERT(a_subIndex<subCount(a_index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									m_attribute!=
									SurfaceAccessibleI::e_properties &&
									!isBound())
							{
								addInteger();
							}

							select(a_index);

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								FEASSERT(m_pPrimitive);
#ifdef FE_HOUDINI_USE_GA
								const GA_Offset offset=
										m_pGdp->pointOffset(a_integer);
								const_cast<GEO_Primitive*>(m_pPrimitive)
										->setPointOffset(a_subIndex,offset);
#else
								const GB_Element* pElement=
										m_pGdp->points()(a_integer);
								const_cast<GEO_Primitive*>(m_pPrimitive)
										->getVertex(a_subIndex).setPt(
										const_cast<GB_Element*>(pElement));
#endif
							}

							//* TODO toggle open curve for e_open

							m_handle.setI(a_integer,a_subIndex);
						}
virtual	I32				integer(U32 a_index,U32 a_subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(a_index<count());
							FEASSERT(m_attribute==
									SurfaceAccessibleI::e_properties ||
									a_subIndex<subCount(a_index));
#endif
							select(a_index);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
#ifdef FE_HOUDINI_USE_GA
								if(!m_pIterator || !m_pIterator->isValid())
								{
									return -1;
								}
								return m_pGdp->pointIndex(
										m_pIterator->getOffset());
#else
								FEASSERT(m_pPoint);
								return m_pPoint->getNum();
#endif
							}

							if(m_attribute==SurfaceAccessibleI::e_properties)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								FEASSERT(m_pPrimitive);
								if(a_subIndex==SurfaceAccessibleI::e_openCurve)
								{
									const GEO_Face* pGeoFace=
											dynamic_cast<const GEO_Face*>(
											m_pPrimitive);

									return pGeoFace && !pGeoFace->isClosed();
								}
								if(a_subIndex==SurfaceAccessibleI::e_countU ||
										a_subIndex==
										SurfaceAccessibleI::e_countV ||
										a_subIndex==
										SurfaceAccessibleI::e_wrappedU ||
										a_subIndex==
										SurfaceAccessibleI::e_wrappedV)
								{
#ifdef FE_HOUDINI_USE_GA
									if(m_pPrimitive->getTypeId().get()!=
											GEO_PRIMMESH)
									{
										return 0;
									}
#else
									if(m_pPrimitive->getPrimitiveId()!=
											GEOPRIMMESH)
									{
										return 0;
									}
#endif
									const GEO_PrimMesh* pPrimMesh=
											dynamic_cast<const GEO_PrimMesh*>(
											m_pPrimitive);
									if(pPrimMesh)
									{
										switch(a_subIndex)
										{
											case SurfaceAccessibleI::e_countU:
												return pPrimMesh->getNumCols();
											case SurfaceAccessibleI::e_countV:
												return pPrimMesh->getNumRows();
											case SurfaceAccessibleI::e_wrappedU:
												return pPrimMesh->isWrappedU();
											case SurfaceAccessibleI::e_wrappedV:
												return pPrimMesh->isWrappedV();
											default:
												;
										}
									}
								}
								return 0;
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								FEASSERT(m_pPrimitive);
#ifdef FE_HOUDINI_USE_GA
								const GA_Offset offset=m_pPrimitive
										->getPointOffset(a_subIndex);
								return m_pGdp->pointIndex(offset);
#else
								return m_pPrimitive->getVertex(a_subIndex)
										.getBasePt()->getNum();
#endif
								return 1;
							}

							return m_handle.getI(a_subIndex);

						}
virtual	I32				duplicate(U32 a_index,U32 a_subIndex=0)
						{
							//* TODO maybe other elements
							if(m_element!=SurfaceAccessibleI::e_point &&
									m_element!=SurfaceAccessibleI::e_primitive)
							{
								return -1;
							}

							FEASSERT(m_pGdpChangeable);
							if(!m_pGdpChangeable)
							{
								return -1;
							}

							I32 dupIndex= -1;
							GEO_Primitive* pDupPrimitive=NULL;

#ifdef FE_HOUDINI_USE_GA
							if(m_element==SurfaceAccessibleI::e_point)
							{
								const GA_Offset offset=
										m_pGdpChangeable->pointOffset(a_index);
								GA_Offset dupOffset=
										m_pGdpChangeable->appendPointCopy(
										offset);
								return m_pGdpChangeable->pointIndex(dupOffset);
							}
							else
							{
								select(a_index);

								FEASSERT(m_pPrimitive);
								const GA_PrimitiveTypeId primType=
										m_pPrimitive->getTypeId();
								pDupPrimitive=
										m_pGdpChangeable->appendPrimitive(
										primType,m_pPrimitive);
								dupIndex=pDupPrimitive->getNum();
							}
#else
							if(m_element==SurfaceAccessibleI::e_point)
							{
								select(a_index);

								FEASSERT(m_pPoint);
								const GEO_Point* pDupPoint=
										m_pGdpChangeable->appendPoint(
										*m_pPoint);
								return pDupPoint->getNum();
							}
							else
							{
								select(a_index);

								FEASSERT(m_pPrimitive);
								const unsigned primType=
										m_pPrimitive->getPrimitiveId();
								pDupPrimitive=
										m_pGdpChangeable->appendPrimitive(
										primType,m_pPrimitive);
								dupIndex=pDupPrimitive->getNum();
							}
#endif

							const GU_PrimPoly* pPolyOrig=
									dynamic_cast<const GU_PrimPoly*>(
									m_pPrimitive);
							GU_PrimPoly* pPolyDup=
									dynamic_cast<GU_PrimPoly*>(pDupPrimitive);

							const U32 vertCount=subCount(a_index);
							for(U32 vertIndex=0;vertIndex<vertCount;
									vertIndex++)
							{
								const I32 pointIndex=integer(a_index,vertIndex);
								append(dupIndex,pointIndex);
							}

							FEASSERT(subCount(dupIndex)==vertCount);

							if(pPolyOrig && pPolyDup && pPolyOrig->isClosed())
							{
								pPolyDup->close();
							}

							return dupIndex;
						}
virtual	I32				append(void)
						{
							if(m_element!=SurfaceAccessibleI::e_primitive)
							{
								return -1;
							}

							//* TODO add new poly
							int npts=0;
							//* TODO GU_POLY_OPEN or GU_POLY_CLOSED
							int open=GU_POLY_CLOSED;
							int appendPts=1;
//							GU_PrimPoly* pGuPrimPoly=
									GU_PrimPoly::build(m_pGdpChangeable,
									npts,open,appendPts);

							return count()-1;
						}
virtual	void			append(U32 a_index,I32 a_integer)
						{
							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup &&
									!(m_element==
									SurfaceAccessibleI::e_primitive &&
									m_attribute==
									SurfaceAccessibleI::e_vertices))
							{
								return;
							}

							FEASSERT(m_pGdpChangeable);
							if(!m_pGdpChangeable)
							{
								return;
							}

							if(m_element==SurfaceAccessibleI::e_primitive)
							{
								select(a_index);

								GEO_TriMesh* pGeoTriMesh=
										dynamic_cast<GEO_TriMesh*>(
										const_cast<GEO_Primitive*>(
										m_pPrimitive));
								FEASSERT(pGeoTriMesh);
#ifdef FE_HOUDINI_USE_GA
								const GA_Offset offset=
										m_pGdpChangeable->pointOffset(
										a_integer);
//								const int vertIndex=
										pGeoTriMesh->appendVertex(offset);
#else
								GEO_Point* pPoint=
										m_pGdpChangeable->points()(a_integer);
//								const int vertIndex=
										pGeoTriMesh->appendVertex(pPoint);
#endif
								return;
							}

							// a_index ignored

							if(!m_pGroup)
							{
								if(m_element==SurfaceAccessibleI::e_pointGroup)
								{
									m_pGroup=m_pGdpChangeable->newPointGroup(
											m_attrName.c_str());
								}
								else
								{
									m_pGroup=
											m_pGdpChangeable->newPrimitiveGroup(
											m_attrName.c_str());
								}
							}

							FEASSERT(m_pGroup);
#ifdef FE_HOUDINI_USE_GA
							m_pGroup->addIndex(a_integer);
#else
							m_pGroup->add(a_integer);
#endif
						}
virtual	void			remove(U32 a_index,I32 a_integer)
						{
							// a_index ignored

							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup)
							{
								return;
							}

							FEASSERT(m_pGdpChangeable);
							if(!m_pGdpChangeable)
							{
								return;
							}

							if(!m_pGroup)
							{
								return;
							}

							FEASSERT(m_pGroup);
#ifdef FE_HOUDINI_USE_GA
							m_pGroup->removeIndex(a_integer);
#else
							m_pGroup->remove(a_integer);
#endif
						}

virtual	void			set(U32 a_index,U32 a_subIndex,Real a_real)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(a_index<count());
							FEASSERT(a_subIndex<subCount(a_index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									m_attribute!=
									SurfaceAccessibleI::e_properties &&
									!isBound())
							{
								addReal();
							}
							select(a_index);

							m_handle.setF(a_real,a_subIndex);
						}
virtual	Real			real(U32 a_index,U32 a_subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(a_index<count());
							FEASSERT(a_subIndex<subCount(a_index));
#endif
							select(a_index);

							if(m_attribute==SurfaceAccessibleI::e_vertices ||
									m_attribute==
									SurfaceAccessibleI::e_properties)
							{
								return 0.0;
							}

							return m_handle.getF(a_subIndex);
						}

virtual	void			set(U32 a_index,U32 a_subIndex,
								const SpatialVector& a_vector)
						{
							if(m_attribute==SurfaceAccessibleI::e_properties)
							{
								return;
							}

							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addSpatialVector();
							}

#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(a_index<count());
							FEASSERT(a_subIndex<subCount(a_index));
#endif

							select(a_index);

							UT_Vector3 utVector3(a_vector[0],
									a_vector[1],a_vector[2]);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
#ifdef FE_HOUDINI_USE_GA
								if(!m_pIterator || !m_pIterator->isValid())
								{
									return;
								}
								m_pGdpChangeable->setPos3(
										m_pIterator->getOffset(),utVector3);
#else
								m_pPointChangeable->setPos(utVector3);
#endif
								return;
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								FEASSERT(m_pPrimitive);
#ifdef FE_HOUDINI_USE_GA
								const GA_Offset offset=
										m_pPrimitive->getPointOffset(a_subIndex);
								m_pGdpChangeable->setPos3(offset,utVector3);
#else
								m_pPrimitive->getVertex(a_subIndex).getPt()
										->setPos(utVector3);
#endif
								return;
							}

							m_handle.setV3(utVector3);
						}
virtual	SpatialVector	spatialVector(U32 a_index,U32 a_subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(a_index<count());
							FEASSERT(a_subIndex<subCount(a_index));
#endif
							if(m_attribute==SurfaceAccessibleI::e_properties)
							{
								return SpatialVector(0.0,0.0,0.0);
							}

							select(a_index);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
#ifdef FE_HOUDINI_USE_GA
								if(!m_pIterator || !m_pIterator->isValid())
								{
									return SpatialVector(0.0,0.0,0.0);
								}
								const UT_Vector3 utVector3=m_pGdp->getPos3(
										m_pIterator->getOffset());
								return SpatialVector(utVector3.data());
#else
								FEASSERT(m_pPoint);
								UT_Vector3 utVector3=m_pPoint->getPos();
								return SpatialVector(utVector3.data());
#endif
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								FEASSERT(m_pPrimitive);
#ifdef FE_HOUDINI_USE_GA
								const GA_Offset offset=
										m_pPrimitive->getPointOffset(a_subIndex);
								const UT_Vector3 utVector3=
										m_pGdp->getPos3(offset);
								return SpatialVector(utVector3.data());
#else
								UT_Vector3 utVector3=m_pPrimitive
										->getVertex(a_subIndex).getPos();
								SpatialVector vector(utVector3.data());
								return vector;
#endif
							}


							UT_Vector3 utVector3=m_handle.getV3(a_subIndex);
							SpatialVector vector(utVector3.data());
							return vector;
						}

						//* Houdini specific
		void			setGdp(const GU_Detail* a_pGdp)
						{
							m_pGdp=a_pGdp;
							m_pGdpChangeable=NULL;
						}
		void			setGdpChangeable(GU_Detail* a_pGdp)
						{
							m_pGdp=a_pGdp;
							m_pGdpChangeable=a_pGdp;
						}
const	GU_Detail*		gdp(void)				{ return m_pGdp; }
		GU_Detail*		gdpChangeable(void)		{ return m_pGdpChangeable; }

		void			harden(void);

	private:
virtual	BWORD			bindInternal(SurfaceAccessibleI::Element a_element,
								const String& a_name);

		void			addString(void);
		void			addInteger(void);
		void			addReal(void);
		void			addSpatialVector(void);

		BWORD			isBound(void) const
						{	return (m_handle.isAttributeValid() || m_pGroup); }

		void			select(U32 a_index);

const	GU_Detail*						m_pGdp;
		GU_Detail*						m_pGdpChangeable;

		GEO_AttributeHandle				m_handle;
const	GEO_Primitive*					m_pPrimitive;
		U32								m_groupIndex;

#ifdef FE_HOUDINI_USE_GA
										//* NOTE may have been cast from const
										//* TODO changeable distinction
		GA_ElementGroup*				m_pGroup;

		GA_Iterator*					m_pIterator;
#else
const	GEO_Point*						m_pPoint;
		GEO_Point*						m_pPointChangeable;
		GB_Group*						m_pGroup;
#endif

#if FE_HOUDINI_HARDENING
	public:

	class Hardening: public Handled<Hardening>
	{
		public:
					Hardening(void):
						m_pHarden(NULL)										{}
					~Hardening(void)
					{	delete m_pHarden; }

			void	harden(GA_AutoHardenForThreading* a_pHarden)
					{	m_pHarden=a_pHarden; }
			BWORD	hardened(void) const
					{	return m_pHarden!=NULL; }

		private:
			GA_AutoHardenForThreading*	m_pHarden;
	};
		sp<Hardening>					m_spHardening;
#endif
};

} // namespace

#endif /* __houdini_SurfaceAccessorHoudini_h__ */
